''' NLP Chatbot Deep Learning '''
import nltk # natural language tool kits
import numpy
import tensorflow as tf
import json
import os

''' 
Preparing a model
'''
def load_data(json_file='intents.json') :
    ''' 
    load intents file 
    intents => { tag, patterns, responses, context_set } 
    ''' 
    with open(json_file) as f:
        dataset = json.load(f)
    
    return dataset
    
    
def tokenize_words(dataset) :        
    words = []
    intents_x = []
    for intent in dataset:
        for p in intent['patterns']+intent['responses'] : # consider both Q & A
            #tokens = nltk.word_tokenize(p)
            tokens = nltk.regexp_tokenize(p, "[\w']+") # can remove ?
            words.extend(tokens)
            intents_x.append(tokens)

    intent_tags = []
    intents_y = []
    for intent in dataset:
        for p in intent['patterns']+intent['responses'] : # consider both Q & A
            intents_y.append(intent["tag"])
        if intent['tag'] not in intent_tags:
            intent_tags.append(intent['tag'])
            
    intent_tags = sorted(intent_tags)

    print("\nwords=")        
    print(words)   
    print("\nintents_x=")  
    print(intents_x) 
    print("\nintents_y=")   
    print(intents_y)
    print("\nintent_tags=")   
    print(intent_tags)

    return (words, intents_x, intents_y, intent_tags)


def stemming(words) :
    '''
    A word stemmer based on the Lancaster (Paice/Husk) stemming algorithm.
    For finding the base or root of words.
    '''
    
    stemmer = nltk.stem.LancasterStemmer()
    words_tmp = []
    for w in words :
        #if w != '?' : words_tmp.append(stemmer.stem(w.lower()))
        words_tmp.append(stemmer.stem(w.lower()))
            
    root_words = sorted(set(words_tmp)) # set() will remove duplications

    print("\nAfter stemming, root words=") 
    print(root_words)    
    
    return root_words


def get_train_data(root_words, intents_x, intents_y, intent_tags) :

    stemmer = nltk.stem.LancasterStemmer()

    training = []
    output = []

    for i in range(len(intents_x)):
        ''' for x '''
        bag_of_words = []
        words_tmp = []
        for w in intents_x[i] :
            #if w != '?' : words_tmp.append(stemmer.stem(w.lower()))
            words_tmp.append(stemmer.stem(w.lower()))        
            
        for w in root_words:
            found = 1 if w in words_tmp else 0
            bag_of_words.append(found)
            
        training.append(bag_of_words)

        ''' for y '''
        output_row = [0] * len(intent_tags)
        output_row[intent_tags.index(intents_y[i])] = 1
        output.append(output_row)

    train_x = numpy.array(training)
    train_y = numpy.array(output)   
    
    print("\n")
    print(train_x.shape)
    print(train_x)
    print(train_y.shape)
    print(train_y)
    
    return (train_x, train_y)


def build_model(train_x, train_y) :
    epochs = 1000
    batch_size = 16

    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Dense(batch_size, input_shape=(len(train_x[0]),))) # input layer
    model.add(tf.keras.layers.Dense(batch_size)) # hidden layer 1
    model.add(tf.keras.layers.Dense(batch_size)) # hidden layer 2
    model.add(tf.keras.layers.Dense(len(train_y[0]), activation='softmax')) # output layer

    model.compile(optimizer="adam", loss="categorical_crossentropy", metrics=["accuracy"])

    model.fit(train_x, train_y, epochs=epochs, batch_size=batch_size, verbose=0)

    ''' Evaluate accuracy of the model '''
    model.summary() # for debugging
    test_loss, test_acc = model.evaluate(train_x, train_y, verbose=0)
    print("\nTested Loss:{} Acc:{}".format(test_loss, test_acc))
    
    return model


def prepare_model(dataset, model_file="model.chatbot") :
    ''' tokenize '''
    words, intents_x, intents_y, intent_tags = tokenize_words(dataset)
    
    ''' stemming '''
    root_words = stemming(words)

    ''' Create a new model '''                 
    train_x, train_y = get_train_data(root_words, intents_x, intents_y, intent_tags)
    model = build_model(train_x, train_y)
    
    ''' save info '''
    model.save(model_file)
    
    with open('root-words.json', 'w', encoding='utf-8') as f :
        json.dump(root_words, f, indent=2, ensure_ascii=False)  
        
    with open('intent-tags.json', 'w', encoding='utf-8') as f :
        json.dump(intent_tags, f, indent=2, ensure_ascii=False)       
    
'''
Start up
'''
prepare_model(load_data())
